//! [`DieselUuid`] is a simple wrapper around [`Uuid`] that can be used in [`diesel`] queries and
//! structs.
//!
//! ## Features
//! - `from_param`: Adds the `rocket::request::FromParam` implementation

use std::{fmt::Display, ops::Deref, str::FromStr};

use diesel::{
    backend::Backend,
    deserialize::{FromSql, FromSqlRow},
    expression::AsExpression,
    mysql::Mysql,
    serialize::ToSql,
    sql_types::Char,
};
use serde::{Deserialize, Serialize};
use uuid::Uuid;

// Re-Export uuid, for convinience.
pub use uuid;

/// Wrapper struct around a [`Uuid`].
///
/// The inner [`Uuid`] can be accessed transparently due to [`DieselUuid`] implementing [`Deref`].
#[derive(
    Debug,
    Clone,
    Copy,
    PartialEq,
    Eq,
    FromSqlRow,
    AsExpression,
    Hash,
    PartialOrd,
    Ord,
    Serialize,
    Deserialize,
)]
#[diesel(sql_type = Char)]
pub struct DieselUuid(Uuid);

impl DieselUuid {
    pub fn new(inner: Uuid) -> Self {
        Self(inner)
    }

    /// Construct a new v4 uuid
    pub fn v4() -> Self {
        Self(Uuid::new_v4())
    }

    pub fn as_uuid(&self) -> &Uuid {
        &self.0
    }

    /// Consume `Self` and return the enclosed [`Uuid`]
    pub fn into_inner(self) -> Uuid {
        self.0
    }
}

impl Display for DieselUuid {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}", self.0.to_simple())
    }
}

impl ToSql<Char, Mysql> for DieselUuid {
    fn to_sql(&self, out: &mut diesel::serialize::Output<Mysql>) -> diesel::serialize::Result {
        <String as ToSql<Char, Mysql>>::to_sql(&self.to_string(), &mut out.reborrow())
    }
}

impl FromSql<Char, Mysql> for DieselUuid {
    fn from_sql(bytes: <Mysql as Backend>::RawValue<'_>) -> diesel::deserialize::Result<Self> {
        let string: String = <String as FromSql<Char, Mysql>>::from_sql(bytes)?;
        let uuid = Uuid::parse_str(&string)?;

        Ok(Self(uuid))
    }
}

impl FromStr for DieselUuid {
    type Err = uuid::Error;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let inner = Uuid::parse_str(s)?;

        Ok(Self(inner))
    }
}

impl Deref for DieselUuid {
    type Target = Uuid;

    fn deref(&self) -> &Self::Target {
        &self.0
    }
}

#[cfg(feature = "rocket")]
mod rocket_impl {
    use super::*;

    use rocket::{
        request::FromParam,
        response::{self, Responder},
        Request,
    };

    impl FromParam<'_> for DieselUuid {
        type Error = uuid::Error;

        fn from_param(param: &str) -> Result<Self, Self::Error> {
            Self::from_str(param)
        }
    }

    impl<'r> Responder<'r, 'static> for DieselUuid {
        fn respond_to(self, req: &'r Request<'_>) -> response::Result<'static> {
            self.to_string().respond_to(req)
        }
    }
}
